# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Library General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

# Copyright 2009 Chris Oliver


import keryx
import libkeryx


class InteractiveCLI:
    commands = {}
    help = "Type 'help' for a list of commands."
    project = None


    def __init__(self):
        # Initialize prompt
        self.set_prompt()

        # Register functions
        functions = {
            'help': [self.display_help, 0, "", "Displays the help message"],
            'create': [self.create_project, 1, 
                     "Usage: create <project name>",
                     "Creates a project"],
            'open': [self.open_project, 1, 
                     "Usage: open <project name>",
                     "Opens a project"],
            'close': [self.close_project, 0, 
                     "Usage: close",
                     "Closes the currently opened project"],
        }
        self.register(functions)


    def register(self, functions):
        """Register functions for projects and cli
        Example Parameters: 
            functions = {
                'create': ['Create', 1],
            }

        This registers the command 'create' to call 'Create' of the project
        and provide it with 1 parameter
        """
        for func, val in functions.items():
            try:            
                properties = {
                    'callback': val[0],
                    'parameters': val[1],
                    'help_msg': val[2],
                    'description': val[3],
                }

                # Actually set the command
                self.commands[func] = properties
            except:
                raise AttributeError, \
                      "Error registering %s: incorrect parameters" % func

    def unregister(self, functions):
        """Unregisters a list of functions"""
        for func in functions:
            del self.commands[func]

    def main(self):
        """Main input loop""" 
        print "Welcome to %s %s\nTo begin, either type 'create <project " \
              "name>' or 'open <project name>' if you have an existing " \
              "project" % (keryx.__appname__.capitalize(), keryx.__version__)
        cmd, params = self.get_input()

        while(not (cmd == 'quit' or cmd == 'exit')):
            self.try_command(cmd, params)

            # Ready for another task
            cmd, params = self.get_input()

        # This makes sure __del__ is called and changes are committed
        self.project = None


    def try_command(self, command, parameters):
        """Tries to execute a command"""
        func, length, err = self.get_func(command)
        if func:
            # Call function only if we have a valid number of parameters
            if length == 0 or len(parameters) == length:
                func(*parameters[0:length])
            elif length == -1:
                func(*parameters[0:])
            else:
                print err
        else:
            print err


    def get_func(self, command):
        """Returns the registered function to call and the number of parameters
        Should only be used after calling check_command
        """
        if not(self.commands.has_key(command)):
            error = "Invalid command '%s'. %s" % (command, self.help)
            return [None, None, error]

        return [self.commands[command]['callback'], 
                self.commands[command]['parameters'],
                self.commands[command]['help_msg']]

    
    def get_input(self):
        """Display prompt and return command and arguments"""
        args = raw_input(self.prompt).split()
        if len(args) > 0:
            return [args[0], args[1:]]
        else:
            return [None, None]


    def set_prompt(self, text=None):
        self.prompt = "%s" % keryx.__appname__
        if text:
            self.prompt += ":%s" % text
        self.prompt += "> "


###############################################################################
# Command callbacks

    def display_help(self):
        print '\n%s %s\n\nCommands:' % (keryx.__appname__, keryx.__version__)

        cmds = self.commands.keys()
        cmds.sort()
        for key in cmds:
            print '    %s - %s' % (key, self.commands[key]['description'])

        print '    exit - Return to the command prompt'
	print


    def create_project(self, name):        
        project = libkeryx.GetProject(name)

        # Verify the project does not exist already
        if project.Exists():
            print "Project %s already exists. Type 'open %s' to open it" % \
                  (name, name)
            return        

        project.Create()

        # Set as open project only if it does not fail to load in any way
        self.project = project
        self.set_prompt(name)
        self.register_extra_functions()

        print "Use 'update' (internet available) or 'updatelocal' (no " \
              "internet available) to get a list of available packages."

    def open_project(self, name):
        project = libkeryx.GetProject(name)
    
        if not(project.Exists()):
            print "Project %s does not exist. Type 'create %s' to create it" \
                    % (name, name)
            return

        #TODO: Install commands?

        # Set as open project only if it does not fail to load in any way
        self.project = project
        self.set_prompt(name)
        self.register_extra_functions()

    def close_project(self):
        #TODO: Uninstall commands
        self.unregister(('updatelocal',
                         'update',
                         'download',
                         'install'))

        self.project = None
        self.set_prompt()

    def register_extra_functions(self):
        # Register functions, must map them to local functions to make sure
        # that __del__ gets called on the current project
        functions = {
            'updatelocal': [self.updatelocal, 0, 
                     "Usage: updatelocal",
                     "Updates the available packages using the local cache"],
            'update': [self.update, 0,
                     "Usage: update",
                     "Updates the available packages from the internet"],
            'download': [self.download, -1,
                     "Usage: download <packages>",
                     "Downloads packages, space separated"],
            'install': [self.install, -1,
                     "Usage: install, or install <packages>",
                     "Installs packages to the offline machine"],
            'listsources': [self.listsources, 0,
                     "Usage: listsources",
                     "Lists software sources in use"],
        }
        self.register(functions)

    def updatelocal(self):
        self.project.UpdateLocal()

    def update(self):
        self.project.Update()

    def download(self, *packages):
        if not packages:
            print "You must provide a package to download"
            return

        self.project.Download(packages)

    def install(self, *packages):
        self.project.Install(packages)

    def listsources(self):
        self.project.ListSources()

