# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Library General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

# Copyright 2009 Chris Oliver

import logging
import sys

try:
    from sqlalchemy import create_engine, ForeignKey, Column, Integer, String
    from sqlalchemy.orm import sessionmaker, relation, backref
    from sqlalchemy.ext.declarative import declarative_base
except:
    logging.error("SQLAlchemy is not installed. Visit www.sqlalchemy.org to " \
                  "get the latest version.")
    sys.exit(1)

Base = declarative_base()
from tables import Project
    
class Definition:
    """Never should be instanciated itself"""
    def __init__(self, project, db_filename):
        # Create the project directores if they do not exist

        # Set the project to be used for other methods in
        # this class
        self.project = project

        # Set the database filename
        sql_filename = "sqlite:///" + db_filename
        self.engine = create_engine(sql_filename)

        # Connect to the database
        self.connection = self.engine.connect()

        # Bind an individual Session to the connection
        Session = sessionmaker()
        self.session = Session(bind=self.connection)
        
        # Create all the tables
        logging.info("Initializing shared database tables...")
        Base.metadata.create_all(self.engine)

        # Call the overridden function of the actual definition
        self.OnInit()

    def __del__(self):
        """Definition destructor
        
        Commits all changes and closes database connection
        """
        print "Cleaning up..."
        self.session.commit()
        self.session.close()

    def Exists(self):
        """Checks project has been created"""
        return self.session.query(Project) \
                    .filter(Project.name==self.project).all()

    #TODO: Use decorator instead of checking Exists(), can we do that?
    def Create(self):
        """Create a new Keryx database"""
        if self.Exists():
            raise AttributeError, "Project already exists"
    
        self.OnCreate()

    def Update(self):
        """Update project from the latest repositories"""
        if not self.Exists():
            raise AttributeError, "Project does not exist"
        self.OnUpdate()
        
    def UpdateLocal(self):
        """Update project from the local cache"""
        if not self.Exists():
            raise AttributeError, "Project does not exist"

        self.OnUpdateLocal()
        
    def Upgrade(self):
        """Download out-of-date packages"""
        if not self.Exists():
            raise AttributeError, "Project does not exist"
        self.OnUpgrade()
        
    def Install(self, package):
        """Install a package"""
        if not self.Exists():
            raise AttributeError, "Project does not exist"
        self.OnInstall(package)
        
    def Depends(self, packages):
        """Return a list of depends/recommends/suggests of packages"""
        if not self.Exists():
            raise AttributeError, "Project does not exist"
        self.OnDependencies(packages)
        
    def Download(self, packages):
        """Download a list of packages"""
        if not self.Exists():
            raise AttributeError, "Project does not exist"
        self.OnDownload(packages)
        
    def Search(self, packages):
        """Print list of packages/descriptions containing provided packages"""
        if not self.Exists():
            raise AttributeError, "Project does not exist"
        self.OnSearch(packages)
        
    def PkgNames(self):
        """Print a list of available packages"""
        if not self.Exists():
            raise AttributeError, "Project does not exist"
        self.OnPkgNames()

    def ListSources(self):
        if not self.Exists():
            raise AttributeError, "Project does not exist"
        self.OnListSources()


###############################################################################
# Overridable functions

    def OnInit(self):
        logging.error("OnInit not implemented yet")
    def OnCreate(self):
        logging.error("OnCreate not implemented yet")
    def OnUpdate(self):
        logging.error("OnUpdate not implemented yet")
    def OnUpdateLocal(self):
        logging.error("OnUpdateLocal not implemented yet")
    def OnUpgrade(self):
        logging.error("OnUpgrade not implemented yet")
    def OnInstall(self, package):
        logging.error("OnInstall not implemented yet")
    def OnDepends(self, packages):
        logging.error("OnDepends not implemented yet")
    def OnDownload(self, packages):
        logging.error("OnDownload not implemented yet")
    def OnSearch(self, packages):
        logging.error("OnSearch not implemented yet")
    def OnPkgNames(self):
        logging.error("OnPkgNames not implemented yet")
    def OnListSources(self):
        logging.error("OnPkgNames not implemented yet")
    
