# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Library General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

# Copyright 2009 Chris Oliver

import commands
import logging
import os
import platform
import re

import keryx
from libkeryx import util
from libkeryx.definition import Definition
from libkeryx.tables import Project

import urlgrabber
from urlgrabber.progress import TextMeter

from database import Install, Package, Relation, Repo, Status
from minideblib import AptRepoClient
from minideblib.DpkgVersion import DpkgVersion

__version__ = '1.0'

class apt_def(Definition):
    apt_client = AptRepoClient.AptRepoClient()

    def OnInit(self):
        # Set the project reference if we are opening an existing project
        try:
            self.project_ref = self.session.query(Project). \
                             filter(Project.name==self.project).one()
        except:
            self.project_ref = None

    def OnCreate(self):
        """Creates the project information in the database"""

        if not(platform.system() == "Linux") or \
            commands.getstatusoutput("aptitude --version")[0] != 0:
            logging.error("Aptitude is not available. Make sure you are " \
                          "running on a Linux system with APT installed.")
            return

        arch = platform.machine()
        if arch == 'i686':
            arch = 'i386'
        elif arch == 'x86_64':
            arch = 'amd64'
        else:
            raise AttributeError, "Unknown architecture %s" % arch

        # Create and store the db project entry
        p = Project(self.project,
                    arch,
                    platform.node(), 
                    "apt_def", 
                    __version__)
        self.session.add(p)

        # Set the newly created project as 
        self.project_ref = p

        # Add the repository db entries and the relationships to the project
        self.__add_repositories()

        # Add the list of installed packages from dpkg status
        self.__add_status("/var/lib/dpkg/status")

    def OnUpdateLocal(self):
        """Update package list from the local cache"""

        #TODO: We actually want this to load packages from database eventually.
        #      Once we have minideblib export to database then we can
        #      implement this. The existing code will be moved to load from
        #      file (load_local) to OnCreate where we can then save to DB.

        # Get project's repo relationships
        rels = self.session.query(Relation). \
                filter(Relation.project_id==self.project_ref.id). \
                filter(Relation.type=="repo").all()
        
        # Get items out of each repo
        repos = [" ".join(self.session.query(Repo). \
                    filter(Repo.id==i.target_id).one().items()) for i in rels]
        
        print "Loading available packages from cache...Please wait."

        self.apt_client._arch = [self.project_ref.architecture]
        self.apt_client.load_local("/var/lib/apt/lists", repos)
        
        print "%i available packages" % \
              len(self.apt_client.get_available_binaries())

    def OnUpdate(self):
        """Update package list from the internet"""
        # Get project's repo relationships
        rels = self.session.query(Relation). \
               filter(Relation.project_id==self.project_ref.id). \
               filter(Relation.type=="repo").all()
        
        # Get items out of each repo
        repos = [" ".join(self.session.query(Repo). \
                 filter(Repo.id==i.target_id).one().items()) for i in rels]

        logging.info("Loading available packages from cache...Please wait.")

        self.apt_client._arch = [self.project_ref.architecture]
        self.apt_client.load_repos(repos)
        
        logging.info("%i available packages" % \
                     len(self.apt_client.get_available_binaries()))

    def OnDownload(self, packages):
        downloads = []
        for p in packages:
            downloads = self.__calculate_downloads(downloads, p)

        #TODO: Test if package deb file is already downloaded (check Relations)
        #      and remove it from the download list if that is the case
        #FIXME: Does this mean we can effectively remove the try/catch around
        #       urlgrabber then?

        if not downloads:
            return

        print "%s will download the following packages:" % \
              keryx.__appname__.capitalize()
        total_size = 0
        for d in downloads:
            print d['package']
            total_size += int(d['size'])
        print "%i new packages will be downloaded\n" \
              "Need to grab %s of archives" % \
              (len(downloads), util.convert_file_size(total_size))
        yesno = raw_input("Do you want to continue [Y/n]? ").lower()

        if yesno and yesno != 'y':
            print "Abort."
            return

        current_dir = os.getcwd()
        folder = os.path.join(os.path.dirname(keryx.__file__), \
                              keryx.dir_download)
        if not os.path.exists(folder):
            os.mkdir(folder)
        os.chdir(folder)

        for d in downloads:
            for url in [url for url in d.get_urls() if url.endswith('.deb')]:
                try:
                    #FIXME: I'm not sure that copy_local is working and we are
                    #       having some serious trouble with TextMeter... :|
                    urlgrabber.urlgrab(url,
                                   copy_local=1,
                                   progress_obj=TextMeter(),
                                   reget='check_timestamp')
                except urlgrabber.grabber.URLGrabError, e:
                    # This is usually thrown when the file is already
                    # downloaded, so just move on
                    pass

        os.chdir(current_dir)

        # Create relations for the packages to be installed
        for p in packages:
            install_rel = Install(p, "downloaded")
            self.project_ref.install.append(install_rel)

        # Export minideblib lists to files here
        #FIXME: We have this here because we don't want to require an 
        #       updatelocal on the offline machine before package installation
        #TODO: Move this to export to the db, and do it when minideblib
        #      is downloading the lists
        lists_folder = os.path.join(folder, 'lists')
        if not os.path.exists(lists_folder):
            os.mkdir(lists_folder)

        print "Exporting lists to files..."
        self.apt_client.to_files(lists_folder, sources=False)

    def OnInstall(self, packages):

        #TODO: Only allow installation on a machine that has the same hostname,
        #      architecture as the original project (pull from DB) and has
        #      aptitude installed

        downloads_folder = os.path.join(os.path.dirname(keryx.__file__), keryx.dir_download)
        # The partial folder must exist in order for apt-get to work
        partial = os.path.join(downloads_folder, 'partial')
        if not os.path.exists(partial):
            os.mkdir(partial)
        lists_folder = os.path.join(downloads_folder, 'lists')

        print "Copying lists over to the local machine..."

        errcode, failure = commands.getstatusoutput("sudo cp %s/* /var/lib/apt/lists" % lists_folder)
        if errcode:
            logging.error(failure)
            return

        print "Updating APT cache..."
        errcode, failure = commands.getstatusoutput("sudo apt-cache gencaches")
        if errcode:
            logging.error(failure)
            return

        # Remove the lock file if it exists
        commands.getoutput("sudo rm %s/lock" % downloads_folder)

        print "Installing new packages...Please wait."
        for p in packages:
            errcode, failure = commands.getstatusoutput("sudo apt-get -y -o dir::cache::archives=\"%s\" install %s" % (downloads_folder, p))
            if errcode:
                print failure
                return
        
        print "Finished installing.\n" \
              "Orange you glad you used %s?" % (keryx.__appname__.capitalize())


    def OnListSources(self):
        """Prints a numbered list of software source lines"""
        # Get project's repo relationships
        rels = self.session.query(Relation). \
                filter(Relation.project_id==self.project_ref.id). \
                filter(Relation.type=="repo").all()
        
        # Get items out of each repo
        repos = [" ".join(self.session.query(Repo). \
                    filter(Repo.id==i.target_id).one().items()) for i in rels]

        # Make a numbered list of repo lines
        repos = ["%i\t%s" % (x, y) for x, y in zip(xrange(len(repos)), repos)]
        print "\n".join(repos)

###############################################################################

    def __add_repositories(self, folder="/etc/apt/"):
        """Add apt list files into database"""
        for root, dirs, files in os.walk(folder):
            
            # Get *.list files
            test = re.compile(".list$", re.IGNORECASE)
            lists = filter(test.search, files)
            
            for file in lists:
                # TODO: Clean up
                # Add line if contains content and doesn't start with #
                lines = [line.strip() for line in \
                         open(os.path.join(root, file), "rb").readlines() \
                         if not line.strip().startswith("#") and line.strip()]

                for line in lines:
                    parts = line.split()
                    type = parts[0]
                    url = parts[1]
                    dist = parts[2]
                    sections = " ".join(parts[3:])

                    # Check for existence
                    found = self.session.query(Repo).filter(Repo.url==url). \
                            filter(Repo.type==type).filter(Repo.url==url). \
                            filter(Repo.dist==dist). \
                            filter(Repo.sections==sections).all()
                    
                    if len(found) > 1:
                        logging.error("Duplicate repository lines in " \
                                      "database for %s" % line)

                    if not found:
                        # Add the new repo
                        repo = Repo(type, url, dist, sections)
                        self.session.add(repo)
                        
                        # Get the id for the newly created repo
                        found = self.session.query(Repo). \
                                filter(Repo.url==url). \
                                filter(Repo.type==type). \
                                filter(Repo.url==url). \
                                filter(Repo.dist==dist). \
                                filter(Repo.sections==sections).all()
                        
                    # Create new relation
                    repo_rel = Relation("repo", found[0].id)
                    self.project_ref.relations.append(repo_rel)

    def __add_status(self, status_file):
        """Add status file entries to Status table"""
        packages = self.__parse_list(status_file)

        # Filter installed packages
        packages = [x for x in packages if x.has_key("Status") \
                    and x["Status"] == "install ok installed"]
        logging.info("%s installed packages" % len(packages))

        # For each installed package
        for pkg in packages:
            #TODO: Check for existing status entry for this package/project
            
            # Add the package
            status = Status(pkg["Package"], pkg["Version"])
            self.project_ref.status.append(status)

    def __parse_list(self, filename):
        """Parse a Packages file list"""
        f = open(filename, "rb")

        packages = []
        pkg = {"Long":"", "Conffiles":""}
        for line in f:
            line = line.strip("\n")
            
            if line.startswith(" "):
                if conf:
                    store = "Conffiles"
                else:
                    store = "Long"
                pkg[store] += unicode("%s\n" % line, errors="ignore")

            elif line.startswith("Conffiles:"):
                conf = True

            elif not line:
                packages.append(pkg)
                pkg = {"Long":"", "Conffiles":""}
            
            else:
                conf = False # Stop collection information for Conffiles
                data = line.split(": ")
                pkg[data[0]] = unicode(data[1:][0], errors="ignore")
                
        f.close()
        return packages

    def __calculate_downloads(self, downloads, package, ver=None):
        """Calculates a list of packages to download"""

        # If we aren't looking for a specific version, get the latest        
        if not ver:
            url, ver = self.apt_client.get_best_binary_version(package)
        
        pkg = self.apt_client.get_binary_name_version(package, ver)
        if pkg:
            pkg = pkg[0]
        else:
            logging.error('Package %s not found.' % package)
            return downloads

        # Add the package itself to be downloaded if it is not installed
        downloads.append(pkg)

        # We need to grab both the depends and recommends packages
        dependencies = []
        if pkg.has_key('depends'):
            dependencies += pkg['depends'].split(',')
        if pkg.has_key('recommends'):
            dependencies += pkg['recommends'].split(',')

        # Iterate through each package
        for dep in dependencies:
            dep = dep.split('|')

            # Build (name, version) tuples for each in this section
            temp = []
            for p in dep:
                t = p.split(' (')
                if len(t) > 1:
                    # Strip spaces and ')' from version
                    temp.append((t[0].strip(), t[1].strip()[:-1]))
                else:
                    temp.append((t[0].strip(), ''))
            dep = temp

            found = False
            # Check if any of these are installed or ready to be downloaded
            #TODO: Check the installed/download package's Provides as well
            for p in dep:
                try:
                    comparison, version = p[1].split()
                    needed_version = DpkgVersion(version)
                except: 
                    # We have no version specified
                    comparison = '>>'
                    needed_version = DpkgVersion('0')

                # Test against status entries
                #FIXME: Can we reference these through the backref install??
                status_refs = self.session.query(Status). \
                                filter(Status.proj_id==self.project_ref.id). \
                                filter(Status.name==p[0]).all()

                for ref in status_refs:
                    installed_version = DpkgVersion(ref.version) 
                    found = self.__compare_found(installed_version, 
                                                 comparison, 
                                                 needed_version)
                    if found:
                        break
   
                if not found:
                    # Test against downloads
                    for d in downloads:
                        if d['package'] == p[0]:
                            # Matching package name, test its version against
                            installed_version = DpkgVersion(d['version'])
                            found = self.__compare_found(installed_version, 
                                                         comparison, 
                                                         needed_version)
                        if found:
                            break

                # We must have found it on the past iteration
                if found: 
                    break

            if not found:
                # Just grab the first package
                #print "Checking %s" % dep[0][0]
                downloads = self.__calculate_downloads(downloads, dep[0][0])
                
        #TODO: For any packages we are upgrading...we need to check the
        #      packages that depend upon those for upgrades as well... :|
        #      Example: samba-common upgrade from 2:3.4.0-3ubuntu5 to
        #               2:3.4.0-3ubuntu5.3 triggers our installed smbclient
        #               package to need an upgrade (along with any others
        #               that might be installed and require a specific version
        #               of a package to be installed). In this case, smbclient
        #               must be the same version number as samba-common

        return downloads

    def __compare_found(self, v1, comparison, v2):
        """Make a comparison between two versions and return the result"""
        if (comparison == '>>' and v1 > v2) or \
           (comparison == '>=' and v1 >= v2) or \
           (comparison == '=' and v1 == v2) or \
           (comparison == '<=' and v1 <= v2) or \
           (comparison == '<<' and v1 < v2): 
            return True
        return False

